package javax.swing.text;

import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.AbstractUndoableEdit;
import java.lang.ref.WeakReference;
import java.util.Vector;
import java.io.IOException;
import java.io.ObjectInputStream;
import javax.swing.undo.UndoableEdit;
import java.lang.ref.ReferenceQueue;
import java.io.Serializable;

public class GapContent extends GapVector implements AbstractDocument.Content, Serializable
{
    private static final char[] empty;
    private transient GapContent.MarkVector marks;
    private transient GapContent.MarkData search;
    private transient int unusedMarks;
    private transient ReferenceQueue<GapContent.StickyPosition> queue;
    static final int GROWTH_SIZE = 524288;
    
    public GapContent() {
        this(10);
    }
    
    public GapContent(int a) {
        super(Math.max(a, 2));
        this.unusedMarks = 0;
        char[] array = { '\n' };
        this.replace(0, 0, array, array.length);
        this.marks = new GapContent.MarkVector();
        this.search = new GapContent.MarkData(0);
        this.queue = new ReferenceQueue();
    }
    
    @Override
    protected Object allocateArray(int n) {
        return new char[n];
    }
    
    @Override
    protected int getArrayLength() {
        return ((char[])(char[])this.getArray()).length;
    }
    
    @Override
    public int length() {
        return this.getArrayLength() - (this.getGapEnd() - this.getGapStart());
    }
    
    @Override
    public UndoableEdit insertString(int n, String s) throws BadLocationException {
        if (n > this.length() || n < 0) {
            throw new BadLocationException("Invalid insert", this.length());
        }
        char[] charArray = s.toCharArray();
        this.replace(n, 0, charArray, charArray.length);
        return new GapContent.InsertUndo(n, s.length());
    }
    
    @Override
    public UndoableEdit remove(int n, int n2) throws BadLocationException {
        if (n + n2 >= this.length()) {
            throw new BadLocationException("Invalid remove", this.length() + 1);
        }
        GapContent.RemoveUndo removeUndo = new GapContent.RemoveUndo(n, this.getString(n, n2));
        this.replace(n, n2, GapContent.empty, 0);
        return removeUndo;
    }
    
    @Override
    public String getString(int n, int n2) throws BadLocationException {
        Segment segment = new Segment();
        this.getChars(n, n2, segment);
        return new String(segment.array, segment.offset, segment.count);
    }
    
    @Override
    public void getChars(int n, int count, Segment segment) throws BadLocationException {
        int n2 = n + count;
        if (n < 0 || n2 < 0) {
            throw new BadLocationException("Invalid location", -1);
        }
        if (n2 > this.length() || n > this.length()) {
            throw new BadLocationException("Invalid location", this.length() + 1);
        }
        int gapStart = this.getGapStart();
        int gapEnd = this.getGapEnd();
        char[] array = (char[])(char[])this.getArray();
        if (n + count <= gapStart) {
            segment.array = array;
            segment.offset = n;
        }
        else if (n >= gapStart) {
            segment.array = array;
            segment.offset = gapEnd + n - gapStart;
        }
        else {
            int count2 = gapStart - n;
            if (segment.isPartialReturn()) {
                segment.array = array;
                segment.offset = n;
                segment.count = count2;
                return;
            }
            segment.array = new char[count];
            segment.offset = 0;
            System.arraycopy(array, n, segment.array, 0, count2);
            System.arraycopy(array, gapEnd, segment.array, count2, count - count2);
        }
        segment.count = count;
        return;
    }
    
    @Override
    public Position createPosition(int n) throws BadLocationException {
        while (this.queue.poll() != null) {
            ++this.unusedMarks;
        }
        if (this.unusedMarks > Math.max(5, this.marks.size() / 10)) {
            this.removeUnusedMarks();
        }
        int gapStart = this.getGapStart();
        int gapEnd = this.getGapEnd();
        int index = (n < gapStart) ? n : (n + (gapEnd - gapStart));
        this.search.index = index;
        int sortIndex = this.findSortIndex(this.search);
        GapContent.MarkData element;
        GapContent.StickyPosition position;
        if (sortIndex < this.marks.size() && (element = this.marks.elementAt(sortIndex)).index == index && (position = element.getPosition()) != null) {
            return position;
        }
        position = new GapContent.StickyPosition();
        GapContent.MarkData mark = new GapContent.MarkData(index, position, this.queue);
        position.setMark(mark);
        this.marks.insertElementAt(mark, sortIndex);
        return position;
    }
    
    @Override
    protected void shiftEnd(int n) {
        int gapEnd = this.getGapEnd();
        super.shiftEnd(n);
        int n2 = this.getGapEnd() - gapEnd;
        int i;
        int size;
        GapContent.MarkData element;
        for (i = this.findMarkAdjustIndex(gapEnd), size = this.marks.size(), ; i < size; ++i) {
            element = this.marks.elementAt(i);
            element.index += n2;
        }
    }
    
    @Override
    int getNewArraySize(int n) {
        if (n < 524288) {
            return super.getNewArraySize(n);
        }
        return n + 524288;
    }
    
    @Override
    protected void shiftGap(int n) {
        int gapStart = this.getGapStart();
        int n2 = n - gapStart;
        int gapEnd = this.getGapEnd();
        int n3 = gapEnd + n2;
        int n4 = gapEnd - gapStart;
        super.shiftGap(n);
        if (n2 > 0) {
            int j = this.findMarkAdjustIndex(gapStart);
            int size2 = this.marks.size();
            while (j < size2) {
                GapContent.MarkData element2 = this.marks.elementAt(j);
                if (element2.index >= n3) {
                    break;
                }
                element2.index -= n4;
                ++j;
            }
        }
        else if (n2 < 0) {
            int i = this.findMarkAdjustIndex(n);
            int size = this.marks.size();
            while (i < size) {
                GapContent.MarkData element = this.marks.elementAt(i);
                if (element.index >= gapEnd) {
                    break;
                }
                element.index += n4;
                ++i;
            }
        }
        this.resetMarksAtZero();
    }
    
    protected void resetMarksAtZero() {
        if (this.marks != null && this.getGapStart() == 0) {
            int gapEnd = this.getGapEnd();
            int i = 0;
            while (i < this.marks.size()) {
                GapContent.MarkData element = this.marks.elementAt(i);
                if (element.index <= gapEnd) {
                    element.index = 0;
                    ++i;
                }
                else {
                    break;
                }
            }
        }
    }
    
    @Override
    protected void shiftGapStartDown(int n) {
        int i = this.findMarkAdjustIndex(n);
        int size = this.marks.size();
        int gapStart = this.getGapStart();
        int gapEnd = this.getGapEnd();
        while (i < size) {
            GapContent.MarkData element = this.marks.elementAt(i);
            if (element.index > gapStart) {
                break;
            }
            element.index = gapEnd;
            ++i;
        }
        super.shiftGapStartDown(n);
        this.resetMarksAtZero();
    }
    
    @Override
    protected void shiftGapEndUp(int index) {
        int i = this.findMarkAdjustIndex(this.getGapEnd());
        int size = this.marks.size();
        while (i < size) {
            GapContent.MarkData element = this.marks.elementAt(i);
            if (element.index >= index) {
                break;
            }
            element.index = index;
            ++i;
        }
        super.shiftGapEndUp(index);
        this.resetMarksAtZero();
    }
    
    final int compare(GapContent.MarkData markData, GapContent.MarkData markData2) {
        if (markData.index < markData2.index) {
            return -1;
        }
        if (markData.index > markData2.index) {
            return 1;
        }
        return 0;
    }
    
    final int findMarkAdjustIndex(int a) {
        this.search.index = Math.max(a, 1);
        int sortIndex = this.findSortIndex(this.search);
        for (int n = sortIndex - 1; n >= 0 && this.marks.elementAt(n).index == this.search.index; --n) {
            --sortIndex;
        }
        return sortIndex;
    }
    
    final int findSortIndex(GapContent.MarkData markData) {
        int i = 0;
        int n = this.marks.size() - 1;
        int n2 = 0;
        if (n == -1) {
            return 0;
        }
        int n3 = this.compare(markData, this.marks.elementAt(n));
        if (n3 > 0) {
            return n + 1;
        }
        while (i <= n) {
            n2 = i + (n - i) / 2;
            n3 = this.compare(markData, this.marks.elementAt(n2));
            if (n3 == 0) {
                return n2;
            }
            if (n3 < 0) {
                n = n2 - 1;
            }
            else {
                i = n2 + 1;
            }
        }
        return (n3 < 0) ? n2 : (n2 + 1);
    }
    
    final void removeUnusedMarks() {
        int size = this.marks.size();
        GapContent.MarkVector marks = new MarkVector(size);
        for (int i = 0; i < size; ++i) {
            GapContent.MarkData element = this.marks.elementAt(i);
            if (element.get() != null) {
                marks.addElement(element);
            }
        }
        this.marks = marks;
        this.unusedMarks = 0;
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws ClassNotFoundException, IOException {
        objectInputStream.defaultReadObject();
        this.marks = new GapContent.MarkVector();
        this.search = new GapContent.MarkData(0);
        this.queue = new ReferenceQueue();
    }
    
    protected Vector getPositionsInRange(Vector vector, int n, int n2) {
        int n3 = n + n2;
        int gapStart = this.getGapStart();
        int gapEnd = this.getGapEnd();
        int i;
        int n4;
        if (n < gapStart) {
            if (n == 0) {
                i = 0;
            }
            else {
                i = this.findMarkAdjustIndex(n);
            }
            if (n3 >= gapStart) {
                n4 = this.findMarkAdjustIndex(n3 + (gapEnd - gapStart) + 1);
            }
            else {
                n4 = this.findMarkAdjustIndex(n3 + 1);
            }
        }
        else {
            i = this.findMarkAdjustIndex(n + (gapEnd - gapStart));
            n4 = this.findMarkAdjustIndex(n3 + (gapEnd - gapStart) + 1);
        }
        vector = ((vector == null) ? new Vector(Math.max(1, n4 - i)) : vector);
        for (; i < n4; ++i) {
            vector.addElement(new GapContent.UndoPosRef(this.marks.elementAt(i)));
        }
        return vector;
    }
    
    protected void updateUndoPositions(Vector vector, int n, int n2) {
        int n3 = n + n2;
        int gapEnd = this.getGapEnd();
        int markAdjustIndex = this.findMarkAdjustIndex(gapEnd + 1);
        int i;
        if (n != 0) {
            i = this.findMarkAdjustIndex(gapEnd);
        }
        else {
            i = 0;
        }
        for (int j = vector.size() - 1; j >= 0; --j) {
            ((GapContent.UndoPosRef)vector.elementAt(j)).resetLocation(n3, gapEnd);
        }
        if (i < markAdjustIndex) {
            Object[] array = new Object[markAdjustIndex - i];
            int n4 = 0;
            if (n == 0) {
                for (; i < markAdjustIndex; ++i) {
                    GapContent.MarkData element3 = this.marks.elementAt(i);
                    if (element3.index == 0) {
                        array[n4++] = element3;
                    }
                }
                for (; i < markAdjustIndex; ++i) {
                    GapContent.MarkData element4 = this.marks.elementAt(i);
                    if (element4.index != 0) {
                        array[n4++] = element4;
                    }
                }
            }
            else {
                for (; i < markAdjustIndex; ++i) {
                    GapContent.MarkData element = this.marks.elementAt(i);
                    if (element.index != gapEnd) {
                        array[n4++] = element;
                    }
                }
                for (; i < markAdjustIndex; ++i) {
                    GapContent.MarkData element2 = this.marks.elementAt(i);
                    if (element2.index == gapEnd) {
                        array[n4++] = element2;
                    }
                }
            }
            this.marks.replaceRange(i, markAdjustIndex, array);
        }
    }
    
    static {
        empty = new char[0];
    }
    
    final class MarkData extends WeakReference<GapContent.StickyPosition>
    {
        int index;
        
        MarkData(int index) {
            super(null);
            this.index = index;
        }
        
        MarkData(int index, GapContent.StickyPosition referent, ReferenceQueue<? super GapContent.StickyPosition> q) {
            super(referent, q);
            this.index = index;
        }
        
        public final int getOffset() {
            int gapStart = GapContent.this.getGapStart();
            int gapEnd = GapContent.this.getGapEnd();
            return Math.max((this.index < gapStart) ? this.index : (this.index - (gapEnd - gapStart)), 0);
        }
        
        GapContent.StickyPosition getPosition() {
            return (GapContent.StickyPosition)this.get();
        }
    }
    
    final class StickyPosition implements Position
    {
        GapContent.MarkData mark;
        
        void setMark(GapContent.MarkData mark) {
            this.mark = mark;
        }
        
        @Override
        public final int getOffset() {
            return this.mark.getOffset();
        }
        
        @Override
        public String toString() {
            return Integer.toString(this.getOffset());
        }
    }
    
    static class MarkVector extends GapVector
    {
        GapContent.MarkData[] oneMark;
        
        MarkVector() {
            this.oneMark = new GapContent.MarkData[1];
        }
        
        MarkVector(int n) {
            super(n);
            this.oneMark = new GapContent.MarkData[1];
        }
        
        @Override
        protected Object allocateArray(int n) {
            return new GapContent.MarkData[n];
        }
        
        @Override
        protected int getArrayLength() {
            return ((GapContent.MarkData[])(GapContent.MarkData[])this.getArray()).length;
        }
        
        public int size() {
            return this.getArrayLength() - (this.getGapEnd() - this.getGapStart());
        }
        
        public void insertElementAt(GapContent.MarkData markData, int n) {
            this.oneMark[0] = markData;
            this.replace(n, 0, this.oneMark, 1);
        }
        
        public void addElement(GapContent.MarkData markData) {
            this.insertElementAt(markData, this.size());
        }
        
        public GapContent.MarkData elementAt(int n) {
            int gapStart = this.getGapStart();
            int gapEnd = this.getGapEnd();
            GapContent.MarkData[] array = (GapContent.MarkData[])(GapContent.MarkData[])this.getArray();
            if (n < gapStart) {
                return array[n];
            }
            n += gapEnd - gapStart;
            return array[n];
        }
        
        protected void replaceRange(int i, int n, Object[] array) {
            int gapStart = this.getGapStart();
            int gapEnd = this.getGapEnd();
            int n2 = 0;
            Object[] array2 = (Object[])(Object[])this.getArray();
            if (i >= gapStart) {
                i += gapEnd - gapStart;
                n += gapEnd - gapStart;
            }
            else if (n >= gapStart) {
                n += gapEnd - gapStart;
                while (i < gapStart) {
                    array2[i++] = array[n2++];
                }
                i = gapEnd;
            }
            else {
                while (i < n) {
                    array2[i++] = array[n2++];
                }
            }
            while (i < n) {
                array2[i++] = array[n2++];
            }
        }
    }
    
    final class UndoPosRef
    {
        protected int undoLocation;
        protected GapContent.MarkData rec;
        
        UndoPosRef(GapContent.MarkData rec) {
            this.rec = rec;
            this.undoLocation = rec.getOffset();
        }
        
        protected void resetLocation(int n, int index) {
            if (this.undoLocation != n) {
                this.rec.index = this.undoLocation;
            }
            else {
                this.rec.index = index;
            }
        }
    }
    
    class InsertUndo extends AbstractUndoableEdit
    {
        protected int offset;
        protected int length;
        protected String string;
        protected Vector posRefs;
        
        protected InsertUndo(int offset, int length) {
            this.offset = offset;
            this.length = length;
        }
        
        @Override
        public void undo() throws CannotUndoException {
            super.undo();
            try {
                this.posRefs = GapContent.this.getPositionsInRange(null, this.offset, this.length);
                this.string = GapContent.this.getString(this.offset, this.length);
                GapContent.this.remove(this.offset, this.length);
            }
            catch (BadLocationException ex) {
                throw new CannotUndoException();
            }
        }
        
        @Override
        public void redo() throws CannotRedoException {
            super.redo();
            try {
                GapContent.this.insertString(this.offset, this.string);
                this.string = null;
                if (this.posRefs != null) {
                    GapContent.this.updateUndoPositions(this.posRefs, this.offset, this.length);
                    this.posRefs = null;
                }
            }
            catch (BadLocationException ex) {
                throw new CannotRedoException();
            }
        }
    }
    
    class RemoveUndo extends AbstractUndoableEdit
    {
        protected int offset;
        protected int length;
        protected String string;
        protected Vector posRefs;
        
        protected RemoveUndo(int offset, String string) {
            this.offset = offset;
            this.string = string;
            this.length = string.length();
            this.posRefs = GapContent.this.getPositionsInRange(null, offset, this.length);
        }
        
        @Override
        public void undo() throws CannotUndoException {
            super.undo();
            try {
                GapContent.this.insertString(this.offset, this.string);
                if (this.posRefs != null) {
                    GapContent.this.updateUndoPositions(this.posRefs, this.offset, this.length);
                    this.posRefs = null;
                }
                this.string = null;
            }
            catch (BadLocationException ex) {
                throw new CannotUndoException();
            }
        }
        
        @Override
        public void redo() throws CannotRedoException {
            super.redo();
            try {
                this.string = GapContent.this.getString(this.offset, this.length);
                this.posRefs = GapContent.this.getPositionsInRange(null, this.offset, this.length);
                GapContent.this.remove(this.offset, this.length);
            }
            catch (BadLocationException ex) {
                throw new CannotRedoException();
            }
        }
    }
}
